with Controller;
with Results;
with Ada.Text_IO;

package body Primes is
   use Common;

   type Index is range 0 .. Common.Search_Limit;

   --  STATIC VARIABLES SECTION

   Holder  : array (Index) of Natural;
   Tail    : Index := 0;
   Current : Index := 1;
   --  Holder records the prime numbers known to this partition
   --  Tail is the index of the latest prime inserted in Holder
   --  Current points to the prime to test againt the number

   --  remember which number we are testing in order to reset Current
   --+ when a new number is tested so that we test it against the values
   --+ in Holder
   Item     : Natural := 0;

   --  attribute 'Partition_ID of unit U returns an integer
   --+ which is the unique id that identifies the partition
   --+ on which unit U is located
   Self : Common.Partition_ID := Primes'Partition_ID;

   --  distributed objects for the current and next partition
   --+ (note that Pool_Access is a class-wide type)
   Own_Pool  : aliased My_Pool_Type;

   --  in order that mutiple instances of this unit may be identified
   --+ by the Main unit that invokes procedure Sieve we register
   --  the distributed object Own_Pool that belongs to the current partition
   --+ to the care of the Controller unit.
   --+ We do so by using a dynamically bound call dispatched according
   --+ to the value of its controlling parameter (Own_Pool'access)
   --+ which is of an access-to-class-wide type
   --+ (this statement is executed during the elaboration of the partition)
   Touch_Base : constant Boolean := Controller.Register (Own_Pool'Access, Self);
   Neighbour  : Controller.Pool_Access := null;

   --  END OF SECTION

   procedure Initialize (Number: in Natural) is
      --  to gain visibility of the operator on Pool_Type
      use Controller;
   begin
      --  we reset Current in case we received a new Number to test for primarity
      if Item /= Number then
         Item := Number;
         Current := 1;
      end if;
      --  and we determine the neighbouring partition
      if Neighbour = null then
         Neighbour := Controller.Next (Self);
      end if;
   end Initialize;

   procedure Sieve
     (Pool     : access My_Pool_Type;
      Number   : in Natural) is
   begin
      Initialize (Number);
      --  we check whether there is any other prime in Holder() to test againt Number
      if Current <= Tail then
         if (Number mod Holder (Current)) = 0 then
            --  since Number is divisible by the current divisor
            --+ we store that divisor in Results.Store
	    Results.Store (Holder (Current), Self);
         else
            --  otherwise Number could be a prime hence we ask our neighboring partition
            --+ to check it against its own set of divisors by making a dynamically-bound
            --+ RPC to that partition
            Sieve (Neighbour, Number);
            --+ as the inquiry can return back to this partition we must increment Current
            Current := Current + 1;
         end if;
      else
	 --  this branch stores Number in Holder
         --+ (this happens at the first invocation of Sieve in the partition
         --+ when Number is certainly a prime
         --+ or when the call to Sieve has returned back to this partition
         --+ and the Number is a prime)
	 Tail := Tail + 1;
         Holder (Tail) := Number;
	 Results.Store (Holder (Current), Self);
      end if;
   end Sieve;

end Primes;
